!
! Copyright (c) 2014-2015 Yoshihide Yoshimoto
! 
program main
  implicit none
  integer,parameter:: nfin1 = 10, nfin2 = 11
  character(len=256):: ifname1, ifname2,argv(2)
  integer:: nkmesh1, nkmesh2, ntb1, ntb2, nocc1, nocc2
  real(kind=8),allocatable:: sklist1(:,:),sklist2(:,:)
  integer,allocatable:: n_band1(:),n_band2(:)
  complex(kind=8),allocatable:: amat1(:),amat2(:)

  real(kind=8),parameter:: eps = 1.0d-5
  character(len=256):: cbuf1, cbuf2
  integer:: nsmpl1,nsmpl2,ndx1(3),ndx2(3),mode1,mode2,nwav1,nwav2
  integer:: nspin1,nspin2,nw1,nw2,itwav1,itwav2,dum
  real(kind=8),allocatable:: aout1(:,:),aout2(:,:)
  integer:: ie,im
  real(kind=8):: ef1,ef2,diffval
  integer:: ic, ib, jb, ik, ios
  logical:: ifdif

  if (iargc().eq.0) then
     write(6,*) 'eps = ',eps
     stop
  else if (iargc().ne.2) then
     write(6,*) 'error in args'
     stop 1
  end if

  ic = 1
  call getarg(ic, argv(ic))
  if (len_trim(argv(ic)) .eq. 0) then
     write(6,*) 'invalid fname1'
     stop 10
  end if
  ifname1 = argv(ic)
  ic = 2
  call getarg(ic, argv(ic))
  if (len_trim(argv(ic)) .eq. 0) then
     write(6,*) 'invalid fname2'
     stop 10
  end if
  ifname2 = argv(ic)

  open(nfin1, file=ifname1, status='old', action='read', form='formatted', iostat=ios)
  if (ios.ne.0) then
     write(6,*) 'can not open fname1'
     stop 11
  end if
  open(nfin2, file=ifname2, status='old', action='read', form='formatted', iostat=ios)
  if (ios.ne.0) then
     write(6,*) 'can not open fname2'
     stop 11
  end if
  rewind nfin1
  rewind nfin2

  do while (.true.)
     read(nfin1,'(a)') cbuf1
     if (trim(cbuf1).eq.' # tetrahedron data') exit
  end do
  read(nfin1,'('' # nsmpl,mode = '',i8,1x,i2)') nsmpl1,mode1
  read(nfin1,'('' # ndx,ndy,ndz ='',3(1x,i6))') ndx1(1),ndx1(2),ndx1(3)
  read(nfin1,'('' # nwav,nw,nspin ='',1x,i3,1x,i6,1x,i2)') nwav1,nw1,nspin1

  do while (.true.)
     read(nfin2,'(a)') cbuf2
     if (trim(cbuf2).eq.' # tetrahedron data') exit
  end do
  read(nfin2,'('' # nsmpl,mode = '',i8,1x,i2)') nsmpl2,mode2
  read(nfin2,'('' # ndx,ndy,ndz ='',3(1x,i6))') ndx2(1),ndx2(2),ndx2(3)
  read(nfin2,'('' # nwav,nw,nspin ='',1x,i3,1x,i6,1x,i2)') nwav2,nw2,nspin2

  rewind(nfin1)
  rewind(nfin2)

  allocate(aout1(0:nwav1,nspin1),aout2(0:nwav2,nspin2))

  if (nsmpl1.ne.nsmpl2) then
     write(6,*) 'number energy sample'
     stop 1
  end if
  if (mode1.ne.mode2) then
     write(6,*) 'tetrahedron mode'
     stop 1
  end if
  if (maxval(abs(ndx2-ndx1)).gt.0) then
     write(6,*) 'dos mesh'
     stop 1
  end if
  if (nw1.ne.nw2) then
     write(6,*) 'number band'
     stop 1
  end if
  if (nwav1.ne.nwav2) then
     write(6,*) 'number wave func'
     stop 1
  end if
  if (nspin1.ne.nspin2) then
     write(6,*) 'number spin'
     stop 1
  end if


  read(nfin1,'()')
  read(nfin2,'()')
  read(nfin1,'()')
  read(nfin2,'()')
  read(nfin1,'()')
  read(nfin2,'()')
  read(nfin1,'(1x,''# itwav, nwav = '',i5,1x,i5)') , itwav1, dum
  read(nfin2,'(1x,''# itwav, nwav = '',i5,1x,i5)') , itwav2, dum

  if (itwav1.ne.itwav2) then
     write(6,*) 'number atomic site'
     stop 1
  end if

  aout1 = 0.0d0
  aout2 = 0.0d0
  do ie=0,nsmpl1
     if (nspin1.eq.1) then
        read(nfin1,*) ef1, (aout1(im,1),im=0,nwav1)
        read(nfin2,*) ef2, (aout2(im,1),im=0,nwav2)
     else
        read(nfin1,*) ef1, (aout1(im,1),im=0,nwav1), &
             (aout1(im,2),im=0,nwav1)
        read(nfin2,*) ef2, (aout2(im,1),im=0,nwav2), &
             (aout2(im,2),im=0,nwav2)
     end if
     if (abs(ef2-ef1).gt.eps) then
        write(6,*) 'energy :', abs(ef2-ef1)
        stop 1
     end if
     diffval = maxval(abs(aout2-aout1))
     if (diffval.gt.eps) then
        write(6,*) 'projected dos :', diffval
        stop 1
     end if
  end do
  read(nfin1,'()')
  read(nfin2,'()')
  if (nspin1.eq.1) then
     read(nfin1,'(a)') cbuf1
     read(cbuf1(3:),*) (aout1(im,1),im=1,nwav1)
     read(nfin2,'(a)') cbuf2
     read(cbuf2(3:),*) (aout2(im,1),im=1,nwav2)
  else
     read(nfin1,'(a)') cbuf1
     read(cbuf1(3:),*) (aout1(im,1),im=1,nwav1), &
          (aout1(im,2),im=1,nwav1)
     read(nfin2,'(a)') cbuf2
     read(cbuf2(3:),*) (aout2(im,1),im=1,nwav2), &
          (aout2(im,2),im=1,nwav2)
  end if
  diffval = maxval(abs(aout2-aout1))
  if (diffval.gt.eps) then
     write(6,*) 'total projected dos :', diffval
     stop 1
  end if

  stop
end program main
