!
! Copyright (c) 2000-2013,2016 Yoshihide Yoshimoto
! 
program main
  implicit none
  integer,parameter:: nfin = 10, nfout = 11
  real(kind=8),parameter:: oneau = 0.529177249
  integer:: nrx21, nry21, nrz21, nrx22, nry22, nrz22
  real(kind=8):: aa1(3,3),aa2(3,3)
  character(len=256):: ifname1, ifname2, argv(4)
  real(kind=8),allocatable:: field1(:,:,:,:), field2(:,:,:,:)
  character(len=8),parameter:: magic = 'TAPP:rho'
  character(len=8):: sign
  integer:: ic
  integer:: nspin1, nspin2
  logical:: ifdif
  real(kind=8):: difval

  real(kind=8),parameter:: epscell = 1.0d-5, epsrho = 1.0d-5

  if (iargc().eq.0) then
     write(6,*) 'eps_cell = ',epscell
     write(6,*) 'eps_rho = ',epsrho
     stop
  else if (iargc().ne.2) then
     write(6,*) 'error in args'
     stop 1
  end if

  ic = 1
  call getarg(ic, argv(ic))
  if (len_trim(argv(ic)) .eq. 0) then
     write(6,*) 'invalid fname1'
     stop
  end if
  ifname1 = argv(ic)
  ic = 2
  call getarg(ic, argv(ic))
  if (len_trim(argv(ic)) .eq. 0) then
     write(6,*) 'invalid fname2'
     stop
  end if
  ifname2 = argv(ic)

  open(nfin, file=ifname1, status='old', action='read', form='unformatted')
  rewind nfin
  read(nfin) sign
  if (sign .ne. magic) then
     write(6,*) 'bad magic for fname1: ', sign
     stop
  end if
  read(nfin) aa1, nspin1
  read(nfin) nrx21, nry21, nrz21
  allocate(field1(nrx21,nry21,nrz21,nspin1))
  read(nfin) field1
  close(nfin)

  open(nfin, file=ifname2, status='old', action='read', form='unformatted')
  rewind nfin
  read(nfin) sign
  if (sign .ne. magic) then
     write(6,*) 'bad magic for fname2: ', sign
     stop
  end if
  read(nfin) aa2, nspin2
  read(nfin) nrx22, nry22, nrz22
  allocate(field2(nrx22,nry22,nrz22,nspin2))
  read(nfin) field2
  close(nfin)

  ifdif = .false.
  if (nrx21.ne.nrx22 .or. nry21.ne.nry22 .or. nrz21.ne.nrz22) then
     ifdif = .true.
     write(6,*) 'cell grid number'
  end if
  if (nspin1.ne.nspin2) then
     ifdif = .true.
     write(6,*) 'number spin'
  end if
  difval = maxval(abs(aa1-aa2))
  if (difval.gt.epscell) then
     ifdif = .true.
     write(6,*) 'cell form', difval
  end if
  difval = maxval(abs(field1-field2))
  if (difval.gt.epsrho) then
     ifdif = .true.
     write(6,*) 'rho field:', difval
  end if
  if (ifdif) then
     stop 1
  end if

  stop
end program main
