!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
! 
module fft_3d
  use pfft3d_const
  use pfft3d_subr
  integer,parameter,private:: nradix=6
  integer,dimension(nradix),parameter,private:: radix=(/6, 8, 3, 4, 2, 5/)
  logical,parameter,private:: prefer4 = .false.
  private:: fact235,genw,ffti,bfti
  !
  type fft3_struct
     integer:: nx,ny,nz,nxyz,lx,ly,lz,lxyz
     integer:: nfx,nfy,nfz
     integer,dimension(nfmax):: facx,facy,facz
     real(kind=8),pointer,dimension(:,:):: wx,wy,wz
     logical:: ifsft
  end type fft3_struct
contains

  subroutine fact235(n,nf,fac)
    implicit none
    integer,intent(in):: n
    integer,intent(out):: nf,fac(nfmax)
    !
    integer:: p,q,r,j
    integer:: n2,n3,n4,n5,n6,n8
    !
    p = n
    fac(:) = 0
    !
    n2 = 0
    q = p/2
    r = p-2*q
    do while ( (q .gt. 0) .and. (r .eq. 0) )
       n2 = n2 + 1
       p = q
       q = p/2
       r = p-2*q
    end do

    n3 = 0
    q = p/3
    r = p-3*q
    do while ( (q .gt. 0) .and. (r .eq. 0) )
       n3 = n3 + 1
       p = q
       q = p/3
       r = p-3*q
    end do

    n5 = 0
    q = p/5
    r = p-5*q
    do while ( (q .gt. 0) .and. (r .eq. 0) )
       n5 = n5 + 1
       p = q
       q = p/5
       r = p-5*q
    end do

    if (p .ne. 1) then
       write(6,*) n, ' can not be factorized with 2, 3, 5'
       stop
    end if

    n6 = min(n2,n3)
    n2 = n2 - n6
    n3 = n3 - n6
    if (prefer4) then ! prefer radix 4
       n4 = n2/2
       j = mod(n2,2)
       if ( (j .eq. 1) .and. (n4 .ge. 1) ) then
          n8 = 1
          n4 = n4 - 1
       else
          n8 = 0
       end if
    else ! prefer radix 8
       n8 = n2/3
       j = mod(n2,3)
       if (j .eq. 2) then
          n4 = 1
       else if ( (j .eq. 1) .and. (n8 .ge. 1) ) then
          n8 = n8 - 1
          n4 = 2
       else
          n4 = 0
       end if
    end if
       
    n2 = n2 - n4*2 - n8*3
    if (n2 .lt. 0) then
       write(6,*) n2, ' n2 become negative'
       stop
    end if
    if (n2+n3+n4+n5+n6+n8 .gt. nfmax) then
       write(6,*) 'too many factors'
       stop
    end if

    p = 1
    nf = 0
    do j=1,n2
       nf = nf + 1
       fac(nf) = 2
       p = p*2
    end do
    do j=1,n3
       nf = nf + 1
       fac(nf) = 3
       p = p*3
    end do
    do j=1,n4
       nf = nf + 1
       fac(nf) = 4
       p = p*4
    end do
    do j=1,n5
       nf = nf + 1
       fac(nf) = 5
       p = p*5
    end do
    do j=1,n6
       nf = nf + 1
       fac(nf) = 6
       p = p*6
    end do
    do j=1,n8
       nf = nf + 1
       fac(nf) = 8
       p = p*8
    end do

    if (p .ne. n) then
       write(6,*) p, ' is not equal to n = ', n
       stop
    end if

    return
  end subroutine fact235

  subroutine genw(n,nf,fac,w)
    implicit none
    integer,intent(in):: n,nf
    integer:: fac(nfmax)
    real(kind=8),intent(out):: w(2,n)
    !
    real(kind=8),parameter:: tpi=2.0d0*3.14159265358979323846d0
    integer:: iw,darg,ir,rad,narg,k,p
    real(kind=8):: argh,argi,arg
    !
    iw = 1
    darg = n
    do ir = 1,nf
       argh = tpi/darg
       rad = fac(ir)
       narg = darg/rad
       do p = 0,narg-1
          do k = 1,rad-1
             argi = argh*k
             arg = argi*p
             w(1,iw) = cos(arg)
             w(2,iw) = -sin(arg)
             iw = iw + 1
          end do
       end do
       darg = darg/rad
    end do
    return
  end subroutine genw
  !
  subroutine ffti(n,nf,fac,w,lx,lyz,dr,di,er,ei,ix)
    implicit none
    integer,intent(in):: n,lx,lyz
    integer,intent(in):: nf,fac(nfmax)
    real(kind=8),intent(in):: w(2,n)
    real(kind=8),dimension(lx,lyz):: dr,di,er,ei
    integer:: ix
    !
    integer:: iw,nvk,np,uif,if,iv
    !
    uif = mod(nf,2)
!poption parallel,tlocal(iw,nvk,np)
    do iv=1,lyz
       iw = 1
       nvk = 1
       np = n/fac(1)
       do if = 1,nf-uif,2
          select case (fac(if))
          case (2)
             call fftrd2(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (3)
             call fftrd3(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (4)
             call fftrd4(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (5)
             call fftrd5(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (6)
             call fftrd6(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (8)
             call fftrd8(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
          iw=iw+np*(fac(if)-1)
          nvk=nvk*fac(if)
          np=np/fac(if+1)
          select case (fac(if+1))
          case (2)
             call fftrd2(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (3)
             call fftrd3(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (4)
             call fftrd4(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (5)
             call fftrd5(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (6)
             call fftrd6(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (8)
             call fftrd8(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
          iw=iw+np*(fac(if+1)-1)
          nvk=nvk*fac(if+1)
          if (if+2 .le. nf) then
             np=np/fac(if+2)
          end if
       end do
       if (uif .eq. 1) then
          select case (fac(nf))
          case (2)
             call fftrd2(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (3)
             call fftrd3(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (4)
             call fftrd4(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (5)
             call fftrd5(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (6)
             call fftrd6(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (8)
             call fftrd8(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
       end if
    end do
    if (uif .eq. 1) then
       ix = - ix
    end if

    return
  end subroutine ffti
  !
  subroutine bfti(n,nf,fac,w,lx,lyz,dr,di,er,ei,ix)
    implicit none
    integer,intent(in):: n,lx,lyz
    integer,intent(in):: nf,fac(nfmax)
    real(kind=8),intent(in):: w(2,n)
    real(kind=8),dimension(lx,lyz):: dr,di,er,ei
    integer:: ix
    !
    integer:: iw,nvk,np,uif,if,iv
    !
    uif = mod(nf,2)
!poption parallel,tlocal(iw,nvk,np)
    do iv=1,lyz
       iw = 1
       nvk = 1
       np = n/fac(1)
       do if = 1,nf-uif,2
          select case (fac(if))
          case (2)
             call bftrd2(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (3)
             call bftrd3(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (4)
             call bftrd4(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (5)
             call bftrd5(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (6)
             call bftrd6(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (8)
             call bftrd8(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
          iw=iw+np*(fac(if)-1)
          nvk=nvk*fac(if)
          np=np/fac(if+1)
          select case(fac(if+1))
          case (2)
             call bftrd2(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (3)
             call bftrd3(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (4)
             call bftrd4(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (5)
             call bftrd5(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (6)
             call bftrd6(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case (8)
             call bftrd8(nvk,np,w(1,iw),er(1,iv),ei(1,iv),dr(1,iv),di(1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
          iw=iw+np*(fac(if+1)-1)
          nvk=nvk*fac(if+1)
          if (if+2 .le. nf) then
             np=np/fac(if+2)
          end if
       end do
       if (uif .eq. 1) then
          select case (fac(nf))
          case (2)
             call bftrd2(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (3)
             call bftrd3(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (4)
             call bftrd4(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (5)
             call bftrd5(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (6)
             call bftrd6(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case (8)
             call bftrd8(nvk,np,w(1,iw),dr(1,iv),di(1,iv),er(1,iv),ei(1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
       end if
    end do
    if (uif .eq. 1) then
       ix = - ix
    end if

    return
  end subroutine bfti
  !
  subroutine fft_pad_zero(nx,ny,nz,lx,ly,lz,dat)
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz
    real(kind=8):: dat(lx,ly,lz,2)
    !
    integer:: j
    if (lz.eq.nz+1) then
       dat(:,:,lz,1) = 0.0d0
       dat(:,:,lz,2) = 0.0d0
    else if (lz.gt.nz+1) then
!poption noparallel
!cdir novector
       do j=nz+1,lz
          dat(:,:,j,1) = 0.0d0
          dat(:,:,j,2) = 0.0d0
       end do
    end if
    if (lx.eq.nx+1) then
       dat(lx,:,:,1) = 0.0d0
       dat(lx,:,:,2) = 0.0d0
    else if (lx.gt.nx+1) then
!poption noparallel
!cdir novector
       do j=nx+1,lx
          dat(j,:,:,1) = 0.0d0
          dat(j,:,:,2) = 0.0d0
       end do
    end if
    if (ly.eq.ny+1) then
       dat(:,ly,:,1) = 0.0d0
       dat(:,ly,:,2) = 0.0d0
    else if (ly.gt.ny+1) then
!poption noparallel
!cdir novector
       do j=ny+1,ly
          dat(:,j,:,1) = 0.0d0
          dat(:,j,:,2) = 0.0d0
       end do
    end if
    return
  end subroutine fft_pad_zero
  !
  subroutine fft3_init(mx,my,mz,lmx,lmy,lmz,fs)
    implicit none
    integer,intent(in):: mx,my,mz,lmx,lmy,lmz
    type (fft3_struct):: fs
    !
    fs%nx = mx
    fs%ny = my
    fs%nz = mz
    fs%nxyz = mx*my*mz
    fs%lx = lmx
    fs%ly = lmy
    fs%lz = lmz
    fs%lxyz = lmx*lmy*lmz
    call fact235(mx,fs%nfx,fs%facx)
    call fact235(my,fs%nfy,fs%facy)
    call fact235(mz,fs%nfz,fs%facz)
    allocate(fs%wx(2,mx),fs%wy(2,my),fs%wz(2,mz))
    call genw(mx,fs%nfx,fs%facx,fs%wx)
    call genw(my,fs%nfy,fs%facy,fs%wy)
    call genw(mz,fs%nfz,fs%facz,fs%wz)
    fs%ifsft = .false.
    return
  end subroutine fft3_init

  subroutine fft3_done(fs)
    implicit none
    type (fft3_struct):: fs
    !
    deallocate(fs%wx,fs%wy,fs%wz)
    return
  end subroutine fft3_done

  subroutine fft3_fw(fs,dat,tmp)
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(fs%lxyz,2)
    real(kind=8):: tmp(fs%lxyz,2)
    integer:: ix


    ix = 1
    call fft3_fwsx(fs,dat,tmp,ix)

    if (ix .eq. 1) then
       continue
    else if (ix .eq. -1) then
       tmp = dat
    else
       write(6,*) 'fft3_fw: f, abs(ix) != 1'
       stop
    end if

    return
  end subroutine fft3_fw

  subroutine fft3_bw(fs,dat,tmp)
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(fs%lxyz,2)
    real(kind=8):: tmp(fs%lxyz,2)
    integer:: ix


    ix = 1
    call fft3_bwsx(fs,dat,tmp,ix)
    !
    if (ix .eq. -1) then
       dat = tmp
    else if (ix .ne. 1) then
       write(6,*) 'fft3_bw: f, abs(ix) != 1'
       stop
    end if
    !
    return
  end subroutine fft3_bw
  !
  subroutine fft3_fwsx(fs,dat,tmp,ix)
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(fs%lxyz,2)
    real(kind=8):: tmp(fs%lxyz,2)
!soption contarea(wx,wy,wz)
    real(kind=8),pointer,dimension(:,:):: wx,wy,wz
    integer:: ix,tix
    integer:: iv,lxy
    real(kind=8):: scl
    !

    wx => fs%wx
    wy => fs%wy
    wz => fs%wz
    call fft_pad_zero(fs%nx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz,tmp)
    !
    if (ix .eq. 1) then
       call ffti(fs%nx, fs%nfx, fs%facx, wx, fs%lx, fs%ly*fs%nz, &
            dat(1,1), dat(1,2), tmp(1,1), tmp(1,2), ix)
    else if (ix .eq. -1) then
       call ffti(fs%nx, fs%nfx, fs%facx, wx, fs%lx, fs%ly*fs%nz, &
            tmp(1,1), tmp(1,2), dat(1,1), dat(1,2), ix)
    else
       write(6,*) 'fft3_fwsx: x, abs(ix) != 1'
       stop
    end if
    !
    lxy = fs%lx*fs%ly
    if (ix .eq. 1) then
!poption parallel,tlocal(tix),lasfv(tix)
!soption nounroll
       do iv=0, fs%nz - 1
          tix = 1
          call fftv(fs%lx, fs%ny, fs%nfy, fs%facy, wy, lxy, &
               dat(iv*lxy+1,1), dat(iv*lxy+1,2), &
               tmp(iv*lxy+1,1), tmp(iv*lxy+1,2), tix)
       end do
       ix = tix
    else if (ix .eq. -1) then
!poption parallel,tlocal(tix),lasfv(tix)
!soption nounroll
       do iv=0, fs%nz - 1
          tix = -1
          call fftv(fs%lx, fs%ny, fs%nfy, fs%facy, wy, lxy, &
               tmp(iv*lxy+1,1), tmp(iv*lxy+1,2), &
               dat(iv*lxy+1,1), dat(iv*lxy+1,2), tix)
       end do
       ix = tix
    else
       write(6,*) 'fft3_fwsx: y, abs(ix) != 1'
       stop
    end if
    !
    if (ix .eq. 1) then
!poption parallel,tlocal(tix),lasfv(tix)
!soption nounroll
       do iv=1, fs%ny
          tix = 1
          call fftvb(fs%lx, fs%ly, iv, fs%nz, fs%nfz, fs%facz, wz, &
               fs%lxyz, dat(1,1), dat(1,2), tmp(1,1), tmp(1,2), tix)
       end do
       ix = tix
    else if (ix .eq. -1) then
!poption parallel,tlocal(tix),lasfv(tix)
!soption nounroll
       do iv=1, fs%ny
          tix = -1
          call fftvb(fs%lx, fs%ly, iv, fs%nz, fs%nfz, fs%facz, wz, &
               fs%lxyz, tmp(1,1), tmp(1,2), dat(1,1), dat(1,2), tix)
       end do
       ix = tix
    else
       write(6,*) 'fft3_fwsx: z, abs(ix) != 1'
       stop
    end if
    !
    scl = 1.0d0/fs%nxyz
    if (ix .eq. 1) then
       dat = scl*dat
    else if (ix .eq. -1) then
       dat = scl*tmp
       ix = 1
    else
       write(6,*) 'fft3_fwsx: s, abs(ix) != 1'
       stop
    end if
    !
    return
  end subroutine fft3_fwsx

  subroutine fft3_bwsx(fs,dat,tmp,ix)
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(fs%lxyz,2)
    real(kind=8):: tmp(fs%lxyz,2)
!soption contarea(wx,wy,wz)
    real(kind=8),pointer,dimension(:,:):: wx,wy,wz
    integer:: ix,tix
    integer:: iv,lxy
    !

    wx => fs%wx
    wy => fs%wy
    wz => fs%wz
    call fft_pad_zero(fs%nx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz,tmp)
    !
    if (ix .eq. 1) then
       call bfti(fs%nx, fs%nfx, fs%facx, wx, fs%lx, fs%ly*fs%nz, &
            dat(1,1), dat(1,2), tmp(1,1), tmp(1,2), ix)
    else if (ix .eq. -1) then
       call bfti(fs%nx, fs%nfx, fs%facx, wx, fs%lx, fs%ly*fs%nz, &
            tmp(1,1), tmp(1,2), dat(1,1), dat(1,2), ix)
    else
       write(6,*) 'fft3_bwsx: x, abs(ix) != 1'
       stop
    end if
    !
    lxy = fs%lx*fs%ly
    if (ix .eq. 1) then
!poption parallel,tlocal(tix),lasfv(tix)
!soption nounroll
       do iv=0, fs%nz - 1
          tix = 1
          call bftv(fs%lx, fs%ny, fs%nfy, fs%facy, wy, lxy, &
               dat(iv*lxy+1,1), dat(iv*lxy+1,2), &
               tmp(iv*lxy+1,1), tmp(iv*lxy+1,2), tix)
       end do
       ix = tix
    else if (ix .eq. -1) then
!poption parallel,tlocal(tix),lasfv(tix)
!soption nounroll
       do iv=0, fs%nz - 1
          tix = -1
          call bftv(fs%lx, fs%ny, fs%nfy, fs%facy, wy, lxy, &
               tmp(iv*lxy+1,1), tmp(iv*lxy+1,2), &
               dat(iv*lxy+1,1), dat(iv*lxy+1,2), tix)
       end do
       ix = tix
    else
       write(6,*) 'fft3_bwsx: y, abs(ix) != 1'
       stop
    end if
    !
    if (ix .eq. 1) then
!poption parallel,tlocal(tix),lasfv(tix)
!soption nounroll
       do iv=1, fs%ny
          tix = 1
          call bftvb(fs%lx, fs%ly, iv, fs%nz, fs%nfz, fs%facz, wz, &
               fs%lxyz, dat(1,1), dat(1,2), tmp(1,1), tmp(1,2), tix)
       end do
       ix = tix
    else if (ix .eq. -1) then
!poption parallel,tlocal(tix),lasfv(tix)
!soption nounroll
       do iv=1, fs%ny
          tix = -1
          call bftvb(fs%lx, fs%ly, iv, fs%nz, fs%nfz, fs%facz, wz, &
               fs%lxyz, tmp(1,1), tmp(1,2), dat(1,1), dat(1,2), tix)
       end do
       ix = tix
    else
       write(6,*) 'fft3_bwsx: z, abs(ix) != 1'
       stop
    end if
    !
    return
  end subroutine fft3_bwsx

end module fft_3d
