!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
! 
module fft_3d
  use pfft3d_const
  use pfft3d_subr
  integer,parameter,private:: nradix=6
  integer,dimension(nradix),parameter,private:: radix=(/6, 8, 3, 4, 2, 5/)
  logical,parameter,private:: prefer4 = .false.
  public:: fft3_init,fft3_fw,fft3_bw,fft3_fwsx,fft3_bwsx,fft3_struct
  !
  type fft3_struct
     integer:: nx,ny,nz,nxyz,lx,ly,lz,lxyz
     integer:: nfx,nfy,nfz
     integer,dimension(nfmax):: facx,facy,facz
     real(kind=8),pointer,dimension(:,:):: wx,wy,wz
     logical:: ifsft
  end type fft3_struct

  integer,parameter,private:: nthread = 4
  integer,parameter,private:: blksz = 1536
contains

  subroutine fact235(n,nf,fac)
    implicit none
    integer,intent(in):: n
    integer,intent(out):: nf,fac(nfmax)
    !
    integer:: p,q,r,j
    integer:: n2,n3,n4,n5,n6,n8
    !
    p = n
    fac(:) = 0
    !
    n2 = 0
    q = p/2
    r = p-2*q
    do while ( (q .gt. 0) .and. (r .eq. 0) )
       n2 = n2 + 1
       p = q
       q = p/2
       r = p-2*q
    end do

    n3 = 0
    q = p/3
    r = p-3*q
    do while ( (q .gt. 0) .and. (r .eq. 0) )
       n3 = n3 + 1
       p = q
       q = p/3
       r = p-3*q
    end do

    n5 = 0
    q = p/5
    r = p-5*q
    do while ( (q .gt. 0) .and. (r .eq. 0) )
       n5 = n5 + 1
       p = q
       q = p/5
       r = p-5*q
    end do

    if (p .ne. 1) then
       write(6,*) n, ' can not be factorized with 2, 3, 5'
       stop
    end if

    n6 = min(n2,n3)
    n2 = n2 - n6
    n3 = n3 - n6
    if (prefer4) then ! prefer radix 4
       n4 = n2/2
       j = mod(n2,2)
       if ( (j .eq. 1) .and. (n4 .ge. 1) ) then
          n8 = 1
          n4 = n4 - 1
       else
          n8 = 0
       end if
    else ! prefer radix 8
       n8 = n2/3
       j = mod(n2,3)
       if (j .eq. 2) then
          n4 = 1
       else if ( (j .eq. 1) .and. (n8 .ge. 1) ) then
          n8 = n8 - 1
          n4 = 2
       else
          n4 = 0
       end if
    end if
       
    n2 = n2 - n4*2 - n8*3
    if (n2 .lt. 0) then
       write(6,*) n2, ' n2 become negative'
       stop
    end if
    if (n2+n3+n4+n5+n6+n8 .gt. nfmax) then
       write(6,*) 'too many factors'
       stop
    end if

    p = 1
    nf = 0
    do j=1,n2
       nf = nf + 1
       fac(nf) = 2
       p = p*2
    end do
    do j=1,n3
       nf = nf + 1
       fac(nf) = 3
       p = p*3
    end do
    do j=1,n4
       nf = nf + 1
       fac(nf) = 4
       p = p*4
    end do
    do j=1,n5
       nf = nf + 1
       fac(nf) = 5
       p = p*5
    end do
    do j=1,n6
       nf = nf + 1
       fac(nf) = 6
       p = p*6
    end do
    do j=1,n8
       nf = nf + 1
       fac(nf) = 8
       p = p*8
    end do

    if (p .ne. n) then
       write(6,*) p, ' is not equal to n = ', n
       stop
    end if

    return
  end subroutine fact235


  subroutine genw(n,nf,fac,w)
    implicit none
    integer,intent(in):: n,nf
    integer:: fac(nfmax)
    real(kind=8),intent(out):: w(2,n)
    !
    real(kind=8),parameter:: tpi=2.0d0*3.14159265358979323846d0
    integer:: iw,darg,ir,rad,narg,k,p
    real(kind=8):: argh,argi,arg
    !
    iw = 1
    darg = n
    do ir = 1,nf
       argh = tpi/darg
       rad = fac(ir)
       narg = darg/rad
       do p = 0,narg-1
          do k = 1,rad-1
             argi = argh*k
             arg = argi*p
             w(1,iw) = cos(arg)
             w(2,iw) = -sin(arg)
             iw = iw + 1
          end do
       end do
       darg = darg/rad
    end do
    return
  end subroutine genw
  !
  subroutine ffti(n,nf,fac,w,lx,lyz,d,e,ix)
    implicit none
    integer,intent(in):: n,lx,lyz
    integer,intent(in):: nf,fac(nfmax)
    real(kind=8),intent(in):: w(2,n)
    real(kind=8),dimension(2,lx,lyz):: d,e
    integer:: ix
    !
    integer:: iw,nvk,np,uif,if,iv
    !
    uif = mod(nf,2)
!poption parallel,tlocal(iw,nvk,np)
    do iv=1,lyz
       iw = 1
       nvk = 1
       np = n/fac(1)
       do if = 1,nf-uif,2
          select case (fac(if))
          case (2)
             call fftrd2(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (3)
             call fftrd3(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (4)
             call fftrd4(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (5)
             call fftrd5(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (6)
             call fftrd6(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (8)
             call fftrd8(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
          iw=iw+np*(fac(if)-1)
          nvk=nvk*fac(if)
          np=np/fac(if+1)
          select case (fac(if+1))
          case (2)
             call fftrd2(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (3)
             call fftrd3(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (4)
             call fftrd4(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (5)
             call fftrd5(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (6)
             call fftrd6(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (8)
             call fftrd8(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
          iw=iw+np*(fac(if+1)-1)
          nvk=nvk*fac(if+1)
          if (if+2 .le. nf) then
             np=np/fac(if+2)
          end if
       end do
       if (uif .eq. 1) then
          select case (fac(nf))
          case (2)
             call fftrd2(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (3)
             call fftrd3(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (4)
             call fftrd4(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (5)
             call fftrd5(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (6)
             call fftrd6(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (8)
             call fftrd8(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
       end if
    end do
    if (uif .eq. 1) then
       ix = - ix
    end if

    return
  end subroutine ffti
  !
  subroutine bfti(n,nf,fac,w,lx,lyz,d,e,ix)
    implicit none
    integer,intent(in):: n,lx,lyz
    integer,intent(in):: nf,fac(nfmax)
    real(kind=8),intent(in):: w(2,n)
    real(kind=8),dimension(2,lx,lyz):: d,e
    integer:: ix
    !
    integer:: iw,nvk,np,uif,if,iv
    !
    uif = mod(nf,2)
!poption parallel,tlocal(iw,nvk,np)
    do iv=1,lyz
       iw = 1
       nvk = 1
       np = n/fac(1)
       do if = 1,nf-uif,2
          select case (fac(if))
          case (2)
             call bftrd2(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (3)
             call bftrd3(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (4)
             call bftrd4(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (5)
             call bftrd5(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (6)
             call bftrd6(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (8)
             call bftrd8(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
          iw=iw+np*(fac(if)-1)
          nvk=nvk*fac(if)
          np=np/fac(if+1)
          select case(fac(if+1))
          case (2)
             call bftrd2(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (3)
             call bftrd3(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (4)
             call bftrd4(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (5)
             call bftrd5(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (6)
             call bftrd6(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case (8)
             call bftrd8(nvk,np,w(1,iw),e(1,1,iv),d(1,1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
          iw=iw+np*(fac(if+1)-1)
          nvk=nvk*fac(if+1)
          if (if+2 .le. nf) then
             np=np/fac(if+2)
          end if
       end do
       if (uif .eq. 1) then
          select case (fac(nf))
          case (2)
             call bftrd2(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (3)
             call bftrd3(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (4)
             call bftrd4(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (5)
             call bftrd5(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (6)
             call bftrd6(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case (8)
             call bftrd8(nvk,np,w(1,iw),d(1,1,iv),e(1,1,iv))
          case default
             write(6,*)'unsupported factor'
             stop
          end select
       end if
    end do
    if (uif .eq. 1) then
       ix = - ix
    end if

    return
  end subroutine bfti
  !
  subroutine fft_pad_zero(nx,ny,nz,lx,ly,lz,dat)
    implicit none
    integer,intent(in):: nx,ny,nz,lx,ly,lz
    real(kind=8):: dat(2,lx,ly,lz)
    !
    integer:: j
    do j=nz+1,lz
       dat(1,:,:,j) = 0.0d0
       dat(2,:,:,j) = 0.0d0
    end do
    do j=nx+1,lx
       dat(1,j,:,:) = 0.0d0
       dat(2,j,:,:) = 0.0d0
    end do
    do j=ny+1,ly
       dat(1,:,j,:) = 0.0d0
       dat(2,:,j,:) = 0.0d0
    end do
    return
  end subroutine fft_pad_zero
  !
  subroutine fft_tp(lxy,lz,src,dst)
    implicit none
    integer,intent(in):: lxy,lz
    real(kind=8),intent(in):: src(2,lxy,lz)
    real(kind=8),intent(out):: dst(2,lz,lxy)
    integer:: i,j
    !
!soption looptiling(64)
    do i=1,lxy
!soption tilesize(64)
       do j=1,lz
          dst(1,j,i) = src(1,i,j)
          dst(2,j,i) = src(2,i,j)
       end do
    end do
    return
  end subroutine fft_tp
  !
  subroutine fft_tps(lxy,lz,src,dst,scl)
    implicit none
    integer,intent(in):: lxy,lz
    real(kind=8),intent(in):: src(2,lxy,lz)
    real(kind=8),intent(out):: dst(2,lz,lxy)
    real(kind=8),intent(in):: scl
    integer:: i,j
    !
!soption looptiling(64)
    do i=1,lxy
!soption tilesize(64)
       do j=1,lz
          dst(1,j,i) = scl*src(1,i,j)
          dst(2,j,i) = scl*src(2,i,j)
       end do
    end do
    return
  end subroutine fft_tps
  !
  subroutine fft3_init(mx,my,mz,lmx,lmy,lmz,fs)
    implicit none
    integer,intent(in):: mx,my,mz,lmx,lmy,lmz
    type (fft3_struct):: fs
    !
    fs%nx = mx
    fs%ny = my
    fs%nz = mz
    fs%nxyz = mx*my*mz
    fs%lx = lmx
    fs%ly = lmy
    fs%lz = lmz
    fs%lxyz = lmx*lmy*lmz
    call fact235(mx,fs%nfx,fs%facx)
    call fact235(my,fs%nfy,fs%facy)
    call fact235(mz,fs%nfz,fs%facz)
    allocate(fs%wx(2,mx),fs%wy(2,my),fs%wz(2,mz))
    call genw(mx,fs%nfx,fs%facx,fs%wx)
    call genw(my,fs%nfy,fs%facy,fs%wy)
    call genw(mz,fs%nfz,fs%facz,fs%wz)
    fs%ifsft = .true.
    return
  end subroutine fft3_init

  subroutine fft3_done(fs)
    implicit none
    type (fft3_struct):: fs
    !
    deallocate(fs%wx,fs%wy,fs%wz)
    return
  end subroutine fft3_done

  subroutine fft3_fw(fs,dat,tmp)
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lxyz)
    real(kind=8):: tmp(2,fs%lxyz)
    integer:: ix


    ix = 1
    call fft3_fwsx(fs,dat,tmp,ix)

    if (ix .eq. 1) then
       call fft_tp(fs%ly*fs%lz, fs%lx, dat, tmp)
       dat = tmp
    else if (ix .eq. -1) then
       call fft_tp(fs%ly*fs%lz, fs%lx, tmp, dat)
    else
       write(6,*) 'fft3_fw: f, abs(ix) != 1'
       stop
    end if
    !
    return
  end subroutine fft3_fw

  subroutine fft3_bw(fs,dat,tmp)
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lxyz)
    real(kind=8):: tmp(2,fs%lxyz)
    integer:: ix


    call fft_tp(fs%lx*fs%ly, fs%lz, dat, tmp)

    ix = 1
    call fft3_bwsx(fs,tmp,dat,ix)

    if (ix .eq. 1) then
       dat = tmp
    else if (ix .eq. -1) then
       continue
    else if (ix .ne. 1) then
       write(6,*) 'fft3_bw: f, abs(ix) != 1'
       stop
    end if

    return
  end subroutine fft3_bw
  !
  subroutine fft3_fwsx(fs,dat,tmp,ix)
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lxyz)
    real(kind=8):: tmp(2,fs%lxyz)
    integer:: ix
    real(kind=8):: scl
    integer:: b,e,tix,nstrd,alcsz

    real(kind=8),allocatable:: bs(:,:),bt(:,:)
!soption contarea(wx,wy,wz)
    real(kind=8),pointer,dimension(:,:):: wx,wy,wz


    wx => fs%wx
    wy => fs%wy
    wz => fs%wz

    alcsz = max(blksz,fs%lz,fs%ly)
    allocate(bs(2,alcsz),bt(2,alcsz))
    !
    if (ix .eq. 1) then
       call fft_pad_zero(fs%nx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz,tmp)
       !
       call ffti(fs%nx, fs%nfx, fs%facx, wx, fs%lx, fs%ly*fs%nz, &
            dat, tmp, ix)
    else if (ix .eq. -1) then
       call fft_pad_zero(fs%nx,fs%ny,fs%nz,fs%lx,fs%ly,fs%lz,dat)
       !
       call ffti(fs%nx, fs%nfx, fs%facx, wx, fs%lx, fs%ly*fs%nz, &
            tmp, dat, ix)
    else
       write(6,*) 'fft3_fwsx: x, abs(ix) != 1'
       stop
    end if
    !
    nstrd = min(blksz/fs%lz,fs%lx*fs%ny/nthread)
    nstrd = max(nstrd,1)

    bs = 0.0d0
    bt = 0.0d0
    if (ix .eq. 1) then
!poption parallel, tlocal(e,tix,bs,bt)
       do b = 1,fs%lx*fs%ny,nstrd
          e = min(b+nstrd-1,fs%lx*fs%ny)

          call fft_tpb(fs%lx*fs%ly, b, e, fs%lz, dat, bs)
          tix = 1
          call fftib(fs%nz, fs%nfz, fs%facz, wz, fs%lz, e-b+1, &
               bs, bt, tix)

          if (tix .eq. 1) then
             call fft_cpb(fs%lx*fs%ly, b, e, fs%lz, bs, tmp)
          else if (tix .eq. -1) then
             call fft_cpb(fs%lx*fs%ly, b, e, fs%lz, bt, tmp)
          else
             write(6,*) 'fft3_fwsx: z, cpb, abs(tix) != 1'
             stop
          end if
       end do
       ix = -1
    else if (ix .eq. -1) then
!poption parallel, tlocal(e,tix,bs,bt)
       do b = 1,fs%lx*fs%ny,nstrd
          e = min(b+nstrd-1,fs%lx*fs%ny)

          call fft_tpb(fs%lx*fs%ly, b, e, fs%lz, tmp, bs)
          tix = 1
          call fftib(fs%nz, fs%nfz, fs%facz, wz, fs%lz, e-b+1, &
               bs, bt, tix)

          if (tix .eq. 1) then
             call fft_cpb(fs%lx*fs%ly, b, e, fs%lz, bs, dat)
          else if (tix .eq. -1) then
             call fft_cpb(fs%lx*fs%ly, b, e, fs%lz, bt, dat)
          else
             write(6,*) 'fft3_fwsx: z, cpb, abs(tix) != 1'
             stop
          end if
       end do
       ix = 1
    else
       write(6,*) 'fft3_fwsx: z, abs(tix) != 1'
       stop
    end if
    !
    nstrd = min(blksz/fs%ly,fs%lz*fs%nx/nthread)
    nstrd = max(nstrd,1)
    scl = 1.0d0/fs%nxyz
    !
    bs = 0.0d0
    bt = 0.0d0
    if (ix .eq. 1) then
!poption parallel, tlocal(e,tix,bs,bt)
       do b = 1,fs%lz*fs%nx,nstrd
          e = min(b+nstrd-1,fs%lz*fs%nx)

          call fft_tpbs(fs%lz*fs%lx, b, e, fs%ly, dat, bs, scl)
          tix = 1
          call fftib(fs%ny, fs%nfy, fs%facy, wy, fs%ly, e-b+1, &
               bs, bt, tix)

          if (tix .eq. 1) then
             call fft_cpb(fs%lz*fs%lx, b, e, fs%ly, bs, tmp)
          else if (tix .eq. -1) then
             call fft_cpb(fs%lz*fs%lx, b, e, fs%ly, bt, tmp)
          else
             write(6,*) 'fft3_fwsx: y, cpb, abs(tix) != 1'
             stop
          end if
       end do
       ix = -1
    else if (ix .eq. -1) then
!poption parallel, tlocal(e,tix,bs,bt)
       do b = 1,fs%lz*fs%nx,nstrd
          e = min(b+nstrd-1,fs%lz*fs%nx)

          call fft_tpbs(fs%lz*fs%lx, b, e, fs%ly, tmp, bs, scl)
          tix = 1
          call fftib(fs%ny, fs%nfy, fs%facy, wy, fs%ly, e-b+1, &
               bs, bt, tix)

          if (tix .eq. 1) then
             call fft_cpb(fs%lz*fs%lx, b, e, fs%ly, bs, dat)
          else if (tix .eq. -1) then
             call fft_cpb(fs%lz*fs%lx, b, e, fs%ly, bt, dat)
          else
             write(6,*) 'fft3_fwsx: y, cpb, abs(tix) != 1'
             stop
          end if
       end do
       ix = 1
    else
       write(6,*) 'fft3_fwsx: y, abs(tix) != 1'
       stop
    end if

    deallocate(bs,bt)
    !
    return
  end subroutine fft3_fwsx

  subroutine fft3_bwsx(fs,dat,tmp,ix)
    implicit none
    type (fft3_struct):: fs
    real(kind=8):: dat(2,fs%lxyz)
    real(kind=8):: tmp(2,fs%lxyz)
    integer:: ix
    integer:: b,e,tix,nstrd,alcsz

    real(kind=8),allocatable:: bs(:,:),bt(:,:)
!soption contarea(wx,wy,wz)
    real(kind=8),pointer,dimension(:,:):: wx,wy,wz


    wx => fs%wx
    wy => fs%wy
    wz => fs%wz

    alcsz = max(blksz,fs%ly,fs%lx)
    allocate(bs(2,alcsz),bt(2,alcsz))
    !
    if (ix .eq. 1) then
       call fft_pad_zero(fs%nz,fs%nx,fs%ny,fs%lz,fs%lx,fs%ly,tmp)
       !
       call bfti(fs%nz, fs%nfz, fs%facz, wz, fs%lz, fs%lx*fs%ny, &
            dat, tmp, ix)
    else if (ix .eq. -1) then
       call fft_pad_zero(fs%nz,fs%nx,fs%ny,fs%lz,fs%lx,fs%ly,dat)
       !
       call bfti(fs%nz, fs%nfz, fs%facz, wz, fs%lz, fs%lx*fs%ny, &
            tmp, dat, ix)
    else
       write(6,*) 'fft3_bwsx: z, abs(ix) != 1'
       stop
    end if
    !
    nstrd = min(blksz/fs%ly,fs%lz*fs%nx/nthread)
    nstrd = max(nstrd,1)

    bs = 0.0d0
    bt = 0.0d0
    if (ix .eq. 1) then
!poption parallel, tlocal(e,tix,bs,bt)
       do b = 1,fs%lz*fs%nx,nstrd
          e = min(b+nstrd-1,fs%lz*fs%nx)

          call fft_tpb(fs%lz*fs%lx, b, e, fs%ly, dat, bs)
          tix = 1
          call bftib(fs%ny, fs%nfy, fs%facy, wy, fs%ly, e-b+1, &
               bs, bt, tix)

          if (tix .eq. 1) then
             call fft_cpb(fs%lz*fs%lx, b, e, fs%ly, bs, tmp)
          else if (tix .eq. -1) then
             call fft_cpb(fs%lz*fs%lx, b, e, fs%ly, bt, tmp)
          else
             write(6,*) 'fft3_bwsx: y, cpb, abs(tix) != 1'
             stop
          end if
       end do
       ix = -1
    else if (ix .eq. -1) then
!poption parallel, tlocal(e,tix,bs,bt)
       do b = 1,fs%lz*fs%nx,nstrd
          e = min(b+nstrd-1,fs%lz*fs%nx)

          call fft_tpb(fs%lz*fs%lx, b, e, fs%ly, tmp, bs)
          tix = 1
          call bftib(fs%ny, fs%nfy, fs%facy, wy, fs%ly, e-b+1, &
               bs, bt, tix)

          if (tix .eq. 1) then
             call fft_cpb(fs%lz*fs%lx, b, e, fs%ly, bs, dat)
          else if (tix .eq. -1) then
             call fft_cpb(fs%lz*fs%lx, b, e, fs%ly, bt, dat)
          else
             write(6,*) 'fft3_bwsx: y, cpb, abs(tix) != 1'
             stop
          end if
       end do
       ix = 1
    else
       write(6,*) 'fft3_bwsx: y, abs(tix) != 1'
       stop
    end if
    !
    nstrd = min(blksz/fs%lx,fs%ly*fs%nz/nthread)
    nstrd = max(nstrd,1)
    !
    bs = 0.0d0
    bt = 0.0d0
    if (ix .eq. 1) then
!poption parallel, tlocal(e,tix,bs,bt)
       do b = 1,fs%ly*fs%nz,nstrd
          e = min(b+nstrd-1,fs%ly*fs%nz)

          call fft_tpb(fs%ly*fs%lz, b, e, fs%lx, dat, bs)
          tix = 1
          call bftib(fs%nx, fs%nfx, fs%facx, wx, fs%lx, e-b+1, &
               bs, bt, tix)

          if (tix .eq. 1) then
             call fft_cpb(fs%ly*fs%lz, b, e, fs%lx, bs, tmp)
          else if (tix .eq. -1) then
             call fft_cpb(fs%ly*fs%lz, b, e, fs%lx, bt, tmp)
          else
             write(6,*) 'fft3_bwsx: x, cpb, abs(tix) != 1'
             stop
          end if
       end do
       ix = -1
    else if (ix .eq. -1) then
!poption parallel, tlocal(e,tix,bs,bt)
       do b = 1,fs%ly*fs%nz,nstrd
          e = min(b+nstrd-1,fs%ly*fs%nz)

          call fft_tpb(fs%ly*fs%lz, b, e, fs%lx, tmp, bs)
          tix = 1
          call bftib(fs%nx, fs%nfx, fs%facx, wx, fs%lx, e-b+1, &
               bs, bt, tix)

          if (tix .eq. 1) then
             call fft_cpb(fs%ly*fs%lz, b, e, fs%lx, bs, dat)
          else if (tix .eq. -1) then
             call fft_cpb(fs%ly*fs%lz, b, e, fs%lx, bt, dat)
          else
             write(6,*) 'fft3_bwsx: x, cpb, abs(tix) != 1'
             stop
          end if
       end do
       ix = 1
    else
       write(6,*) 'fft3_bwsx: x, abs(tix) != 1'
       stop
    end if

    deallocate(bs,bt)
    !
    return
  end subroutine fft3_bwsx

end module fft_3d
