!
! Copyright (c) 2000-2013,2016 Yoshihide Yoshimoto
! 
! argument: ix
!   The original unit vectors in real space a_i is
!   transformed to expanded unit vectors ax_i as
!
!   (ax_1,ax_2,ax_3) = (a_1,a_2,a_3)*ix.
!
!   All element of ix are integer.
!
!   Therefore original unit vectors in the reciprocal space b_i is
!   transformed to expanded unit vectors bx_i as
!
!   (b_1,b_2,b_3) = (bx_1,bx_2,bx_3)*(ix)^t
!
#include "fftgrid.h"

program wfk2dx
  use fft_3d
  implicit none
  integer, parameter:: nfin = 10, nfout = 11

  character(len=8):: magic = 'TAPP:wfk'
  real(kind=8),parameter:: oneau = 0.529177249d0

  integer,parameter:: iorg(3) = (/ 0, 0, 0 /)
  character(len=8):: sign
  integer:: nwx2, nwy2, nwz2, nkt, nbba1, nbba2, maxr, lmax, nsi, nkd
  integer:: nb1, nb2, nkt1, nkt2, nspin, ncomp
  real(kind=8),allocatable:: wfn(:,:,:,:,:),tmp(:,:,:,:),wfnpr(:,:,:,:,:)
  real(kind=8):: aa(3,3), aax(3,3), sk(3), rlat(3), rlp, imgp, ph
  real(kind=8):: delta(3,3)
  real(kind=8):: rorg(3), rorgaa(3), bbx(3,3), w, gg(3)
  integer:: ix(3,3), mx, my, mz
  real(kind=8):: norm, nrmmax, wfmax(2)
  real(kind=8),parameter:: pi = 3.1415926535897932385d0
  real(kind=8),allocatable:: ee(:,:)
  integer,allocatable:: nl(:),nr(:,:),nprj(:),kd(:)
  integer:: laprj
  integer:: ik, ib, ig, i1, i2, i3, k1, k2, k3, j1, j2, j3
  integer:: ir, im, il, it, nlx, nly, nlz, nkm
  integer:: ikd, is, isc
  integer:: ip
  integer,allocatable:: kg(:,:)
  complex(kind=8),allocatable:: p(:,:,:,:)
  complex(kind=8),allocatable:: pby(:,:,:,:)
  character(len=256):: ifname, ofname, name, buf1, buf2, argv(10)
  integer:: ic, ios
  logical:: phaseq, prefrealq, expandq, error, partnbq, partnktq
  type(fft3_struct):: fs
  real(kind=8):: eebuf(2)
  complex(kind=8):: ctemp

  ic = 1
  phaseq = .false.
  prefrealq = .false.
  call getarg(ic, argv(ic))
  if ( argv(ic) .eq. '-p' ) then
     phaseq = .true.
     ic = ic + 1
  end if
  call getarg(ic, argv(ic))
  if ( argv(ic) .eq. '-r' ) then
     prefrealq = .true.
     ic = ic + 1
  end if
  call getarg(ic, argv(ic))
  if (len_trim(argv(ic)) .eq. 0) then
     write(6,*) 'invalid fname'
     stop
  end if
  ifname = argv(ic)
  ic = ic + 1
  call getarg(ic, argv(ic))
  read(argv(ic),*,iostat=ios) ix(1,1), ix(2,1), ix(3,1)
  if (ios .eq. 0) then
     error = .false.
     ic = ic + 1
     call getarg(ic, argv(ic))
     read(argv(ic),*,iostat=ios)  ix(1,2), ix(2,2), ix(3,2)
     error = error .or. (ios .ne. 0)
     ic = ic + 1
     call getarg(ic, argv(ic))
     read(argv(ic),*,iostat=ios)  ix(1,3), ix(2,3), ix(3,3)
     error = error .or. (ios .ne. 0)
     ic = ic + 1
     call getarg(ic, argv(ic))
     read(argv(ic),*,iostat=ios)  rorg(1), rorg(2), rorg(3)
     error = error .or. (ios .ne. 0)
     ic = ic + 1
     call getarg(ic, argv(ic))
     read(argv(ic),*,iostat=ios)  mx, my, mz
     error = error .or. (ios .ne. 0)
     if (error) then
        write(6,*) 'invalid position of origin'
        stop
     end if
     expandq = .true.
     write(6,'(1x,a)') 'expanding cell'
  else
     ix = 0
     ix(1,1) = 1
     ix(2,2) = 1
     ix(3,3) = 1
     rorg = 0.0d0
     expandq = .false.
     write(6,'(1x,a)') 'unit cell'
  end if
  ic = ic + 1
  call getarg(ic, argv(ic))
  read(argv(ic),*,iostat=ios) nb1, nb2
  if (ios .eq. 0) then
     partnbq = .true.
     write(6,'(1x,a,2(1x,i4))') 'nb1, nb2 =', nb1, nb2
  else
     partnbq = .false.
  end if
  ic = ic + 1
  call getarg(ic, argv(ic))
  read(argv(ic),*,iostat=ios) nkt1, nkt2
  if (ios .eq. 0) then
     partnktq = .true.
     write(6,'(1x,a,2(1x,i4))') 'nkt1, nkt2 =', nkt1, nkt2
  else
     partnktq = .false.
  end if

  if ((.not. phaseq) .and. prefrealq) then
     write(6,'(1x,a)') '-r is not supported for wfk only mode'
     stop
  end if

  open(nfin, file=ifname, status='old', action='read', form='unformatted')
  read(nfin) sign
  if (sign .ne. magic) then
     write(6,*) 'Bad magic'
     stop
  end if
  read(nfin,iostat=ios) aa,nspin,ncomp
  if (ios.ne.0) then
     ncomp = 1
     backspace(nfin)
     read(nfin) aa,nspin
  end if
  aax = matmul(aa,ix)
  if (nspin*ncomp.ne.1 .and. nspin*ncomp.ne.2) then
     write(6,*) 'invalid number_spin and number_component in wfn file'
     stop
  end if

  w=aax(1,1)*(aax(2,2)*aax(3,3)-aax(3,2)*aax(2,3)) &
       +aax(2,1)*(aax(3,2)*aax(1,3)-aax(1,2)*aax(3,3)) &
       +aax(3,1)*(aax(1,2)*aax(2,3)-aax(2,2)*aax(1,3))
  bbx(1,1)=(aax(2,2)*aax(3,3)-aax(3,2)*aax(2,3))/w
  bbx(2,1)=(aax(3,2)*aax(1,3)-aax(1,2)*aax(3,3))/w
  bbx(3,1)=(aax(1,2)*aax(2,3)-aax(2,2)*aax(1,3))/w
  bbx(1,2)=(aax(2,3)*aax(3,1)-aax(3,3)*aax(2,1))/w
  bbx(2,2)=(aax(3,3)*aax(1,1)-aax(1,3)*aax(3,1))/w
  bbx(3,2)=(aax(1,3)*aax(2,1)-aax(2,3)*aax(1,1))/w
  bbx(1,3)=(aax(2,1)*aax(3,2)-aax(3,1)*aax(2,2))/w
  bbx(2,3)=(aax(3,1)*aax(1,2)-aax(1,1)*aax(3,2))/w
  bbx(3,3)=(aax(1,1)*aax(2,2)-aax(2,1)*aax(1,2))/w

  read(nfin) nwx2,nwy2,nwz2
  write(6,'(1x,a,3(1x,i4))') 'nw[xyz]2 =',nwx2,nwy2,nwz2
  if (expandq) then
     nwx2 = mx
     nwy2 = my
     nwz2 = mz
  end if

  delta(:,1) = oneau*aax(:,1)/nwx2
  delta(:,2) = oneau*aax(:,2)/nwy2
  delta(:,3) = oneau*aax(:,3)/nwz2
  rorgaa = oneau*rorg

  read(nfin) nkt,nbba1,nbba2
  if (.not. partnbq) then
     nb1 = nbba1
     nb2 = nbba2
  end if
  if (.not. partnktq) then
     nkt1 = 1
     nkt2 = nkt
  end if
  read(nfin) nsi,nkd
  allocate(nl(nkd))
  read(nfin) (nl(ikd),ikd=1,nkd)
  lmax = maxval(nl)
  allocate(nr(lmax,nkd),kd(nsi))
  read(nfin) ((nr(il,ikd),il=1,nl(ikd)),ikd=1,nkd)
  maxr = maxval(nr)
  read(nfin) kd
  read(nfin) laprj
  allocate(nprj(nkd))
  read(nfin) nprj
  allocate(ee(nbba1:nbba2,nspin))

  nlx = FFT_PAD_NX(nwx2)
  nly = FFT_PAD_NY(nwy2)
  nlz = FFT_PAD_NZ(nwz2)
  call fft3_init(nwx2,nwy2,nwz2,nlx,nly,nlz,fs)
     
  allocate(FFT_A3M_DEFSZ(wfn,nlx,nly,nlz,nspin*ncomp))
  allocate(FFT_A3D_DEFSZ(tmp,nlx,nly,nlz))
  allocate(pby(laprj,nsi,nbba1:nbba2,nspin))

  do ik=1,nkt
     read(nfin) sk,nkm,ee
     allocate(kg(3,nkm),p(ncomp,nkm,nbba1:nbba2,nspin))
     read(nfin) kg
     do is=1,nspin
        do ib=nbba1,nbba2
           read(nfin) p(:,:,ib,is)
        end do
     end do
     read(nfin) ((((pby(ip,it,ib,is),ip=1,nprj(kd(it))),it=1,nsi) &
          ,ib=nbba1,nbba2),is=1,nspin)

     if ( (ik.lt.nkt1) .or. (nkt2.lt.ik) ) then
        deallocate (kg,p)
        cycle
     end if

     do ib=nb1,nb2
        write(buf1,'(i0)') ib
        buf1=adjustl(buf1)
        write(buf2,'(i0)') ik
        buf2=adjustl(buf2)

        wfn = 0.0d0
        isc = 0
        do is=1,nspin
           do ic=1,ncomp
              isc = isc + 1
              do ig=1,nkm
                 i1 = kg(1,ig)
                 i2 = kg(2,ig)
                 i3 = kg(3,ig)
                 j1 = ix(1,1)*i1+ix(2,1)*i2+ix(3,1)*i3
                 j2 = ix(1,2)*i1+ix(2,2)*i2+ix(3,2)*i3
                 j3 = ix(1,3)*i1+ix(2,3)*i2+ix(3,3)*i3
                 k1 = mod(j1+nwx2,nwx2)+1
                 k2 = mod(j2+nwy2,nwy2)+1
                 k3 = mod(j3+nwz2,nwz2)+1
                 gg(1) = bbx(1,1)*j1+bbx(1,2)*j2+bbx(1,3)*j3
                 gg(2) = bbx(2,1)*j1+bbx(2,2)*j2+bbx(2,3)*j3
                 gg(3) = bbx(3,1)*j1+bbx(3,2)*j2+bbx(3,3)*j3
                 ph = 2.0d0*pi*(gg(1)*rorg(1)+gg(2)*rorg(2)+gg(3)*rorg(3))
                 ctemp = p(ic,ig,ib,is)*cmplx(cos(ph),sin(ph))
                 FFT_A3M_RL(wfn,k1,k2,k3,isc) = dble(ctemp)
                 FFT_A3M_IM(wfn,k1,k2,k3,isc) = aimag(ctemp)
              end do

              call fft3_bw(fs, FFT_A3M_RL(wfn,1,1,1,isc), tmp)
           end do
        end do

        ofname=trim(ifname)//'.'//trim(buf1)//'.'//trim(buf2)//'.dx' &
             //achar(0)

        if (phaseq) then
           allocate(wfnpr(0:nwx2,0:nwy2,0:nwz2,2,nspin*ncomp))
           do isc=1,nspin*ncomp
              do i3=0,nwz2
                 do i2=0,nwy2
                    do i1=0,nwx2
                       rlat(1) = ix(1,1)*dble(i1)/nwx2 &
                            + ix(1,2)*dble(i2)/nwy2 + ix(1,3)*dble(i3)/nwz2 &
                            + rorg(1)
                       rlat(2) = ix(2,1)*dble(i1)/nwx2 &
                            + ix(2,2)*dble(i2)/nwy2 + ix(2,3)*dble(i3)/nwz2 &
                            + rorg(2)
                       rlat(3) = ix(3,1)*dble(i1)/nwx2 &
                            + ix(3,2)*dble(i2)/nwy2 + ix(3,3)*dble(i3)/nwz2 &
                            + rorg(3)
                       ph = 2.0d0*pi*(sk(1)*rlat(1) + sk(2)*rlat(2) &
                            + sk(3)*rlat(3))
                       rlp = FFT_A3M_RL(wfn,mod(i1,nwx2)+1,mod(i2,nwy2)+1,mod(i3,nwz2)+1,isc)
                       imgp = FFT_A3M_IM(wfn,mod(i1,nwx2)+1,mod(i2,nwy2)+1,mod(i3,nwz2)+1,isc)
                       wfnpr(i1,i2,i3,1,isc) = rlp*cos(ph) - imgp*sin(ph)
                       wfnpr(i1,i2,i3,2,isc) = rlp*sin(ph) + imgp*cos(ph)
                    end do
                 end do
              end do
           end do

           if (prefrealq) then
              do is=1,nspin
                 nrmmax = 0.0d0
                 wfmax = 0.0d0
                 do i3=0,nwz2
                    do i2=0,nwy2
                       do i1=0,nwx2
                          do ic=1,ncomp
                             isc = ic*is
                             norm = wfnpr(i1,i2,i3,1,isc)*wfnpr(i1,i2,i3,1,isc) &
                                  + wfnpr(i1,i2,i3,2,isc)*wfnpr(i1,i2,i3,2,isc)
                             if (norm.gt.nrmmax) then
                                nrmmax = norm
                                wfmax(1) = wfnpr(i1,i2,i3,1,isc)
                                wfmax(2) = wfnpr(i1,i2,i3,2,isc)
                             end if
                          end do
                       end do
                    end do
                 end do
                 write(6,'(3(1x,i4))') ik,ib,is
                 write(6,'(1x,a,e18.11)') 'nrmmax = ', nrmmax
                 norm = wfmax(1)*wfmax(1)+wfmax(2)*wfmax(2)
                 write(6,'(1x,a,e18.11)') 'norm = ', norm
                 norm = sqrt(norm)
                 wfmax = wfmax/norm
                 write(6,'(1x,a,2(1x,e18.11))') 'wfmax =', wfmax(1), wfmax(2)
                 do ic=1,ncomp
                    isc = ic*is
                    do i3=0,nwz2
                       do i2=0,nwy2
                          do i1=0,nwx2
                             rlp = wfnpr(i1,i2,i3,1,isc)
                             imgp = wfnpr(i1,i2,i3,2,isc)
                             wfnpr(i1,i2,i3,1,isc) = rlp*wfmax(1) + imgp*wfmax(2)
                             wfnpr(i1,i2,i3,2,isc) = imgp*wfmax(1) - rlp*wfmax(2)
                          end do
                       end do
                    end do
                 end do
              end do
           end if
              
           name='TAPP exp(ikr)*wfk'//achar(0)

           eebuf = ee(ib,1)
           if (nspin.eq.2) eebuf(2) = ee(ib,2)
           call outzfldprsk(ofname, name, nwx2+1, nwy2+1, nwz2+1, delta, &
                rorgaa, nspin*ncomp, nwx2+1, nwy2+1, nwz2+1, wfnpr, 0, &
                sk, eebuf, 0, iorg)

           deallocate(wfnpr)
        else
           allocate(wfnpr(nlx,nly,nlz,2,nspin*ncomp))
           wfnpr(:,:,:,1,:) = FFT_A3M_RL(wfn,:,:,:,:)
           wfnpr(:,:,:,2,:) = FFT_A3M_IM(wfn,:,:,:,:)

           name='TAPP wfk'//achar(0)

           eebuf = ee(ib,1)
           if (nspin.eq.2) eebuf(2) = ee(ib,2)
           call outzfldprsk(ofname, name, nwx2, nwy2, nwz2, delta, &
                rorgaa, nspin*ncomp, nlx, nly, nlz, wfnpr, 0, &
                sk, eebuf, 1, iorg)

           deallocate(wfnpr)
        end if
     end do
     deallocate(kg,p)
  end do
  deallocate(wfn,tmp,pby)
  deallocate(nl,nr,kd,nprj,ee)
end program wfk2dx
