!
! Copyright (c) 2014,2016 Yoshihide Yoshimoto
! 
program wfneig
  implicit none
  integer, parameter:: nfin = 10

  character(len=8):: sign
  integer:: ver
  integer:: nks, nw, ng, nnp
  integer:: nkmmax
  integer,allocatable:: nkmary(:), kg(:,:), rg(:,:,:), pg(:,:)
  real(kind=8),allocatable:: sk(:,:)
  real(kind=8),allocatable:: pr(:)
  complex(kind=8),allocatable:: p(:,:)
  integer:: nspin, ncomp
  real(kind=8):: bb(3,3), aa(3,3)
  integer:: i, j, ios, ik, is, ib, npbb, npbe
  character(len=256):: ifname, argv(4)
  logical:: error,ifinvsymi
  real(kind=8),allocatable:: ww(:,:), wsum(:,:)
  real(kind=8),allocatable:: ee(:,:), eemax(:,:), eemin(:,:)
  real(kind=8),parameter:: tohart = 27.211396162d0

  call getarg(1, argv(1))
  if (len_trim(argv(1)) .eq. 0) then
     write(6,*) 'invalid fname'
     stop
  end if
  ifname = argv(1)
  call getarg(2, argv(2))
  read(argv(2),*,iostat=ios) npbb
  call getarg(3, argv(3))
  read(argv(3),*,iostat=ios) npbe

  open(nfin, file=ifname, status='old', action='read', form='unformatted')
  ifinvsymi = .false.
  read(nfin, iostat=ios) sign, ver
  if (ios.ne.0) then
     write(6,*) 'unknown format'
     stop
  end if
  if (ver.ne.3) then
     write(6,*) 'only version 3 format is supported by this program'
     stop
  end if
  if (sign .ne. 'TAPP:wfn') then
     if (sign .ne. 'TAPP:wfi') then
        write(6,*) 'Bad magic'
        stop
     else
        ifinvsymi = .true.
     end if
  end if
  read(nfin) aa, bb
  read(nfin) nspin, nks, nw, ng, nnp, ncomp
  allocate(nkmary(nks), sk(3,nks), rg(3,3,ng), pg(3,ng))
  read(nfin) rg, pg
  read(nfin) (sk(:,ik),ik=1,nks)
  read(nfin) nkmary
  nkmmax=maxval(nkmary)
  allocate(pr(nkmmax),p(ncomp,nkmmax),ww(nw,nspin),wsum(nw,nspin), &
       ee(nw,nspin),eemin(nw,nspin),eemax(nw,nspin))

  if (npbb.eq.0) npbb = 1
  if (npbe.eq.0) npbe = nw

  if (npbb.lt.0 .or. npbb.gt.nw &
       .or. npbe.lt.0 .or. npbe.gt.nw .or. npbb .gt. npbe ) then
     write(6,*) 'invalid band range'
     stop
  end if

  write(6,'(1x,a,i5)') '# number k point = ', nks
  write(6,'(1x,a,i5)') '# number band = ', nw
  write(6,'(1x,a,i5,1x,i5)') '# band range = ', npbb, npbe
  write(6,'(1x,a)') '# index, each_spin[ weight, eigen energy(eV) ]'
  write(6,*)

  eemax = -huge(1.0d0)
  eemin = huge(1.0d0)
  wsum = 0.0d0

  do ik=1,nks
     write(6,'(1x,a,i5,a)') '# ', ik, ' -th k point'
     write(6,'(1x,a,f9.6,1x,f9.6,1x,f9.6)') &
          '# sk = ', sk(1,ik), sk(2,ik), sk(3,ik)

    ! read(nfin) ((kg(i,ig),i=1,3),ig=1,nkmary(ik))
     read(nfin)

     do is=1,nspin 
        do ib=1,nw
           if (ifinvsymi) then
              read(nfin) (pr(i),i=1,nkmary(ik)), ww(ib,is), ee(ib,is)
           else
              read(nfin) ((p(i,j),i=1,ncomp),j=1,nkmary(ik)), &
                   ww(ib,is), ee(ib,is)
           end if
           eemax(ib,is) = max(eemax(ib,is),ee(ib,is))
           eemin(ib,is) = min(eemin(ib,is),ee(ib,is))
           wsum(ib,is) = wsum(ib,is) + ww(ib,is)
        end do
     end do

     if (nspin.eq.1) then
        do ib=npbb,npbe
           write(6,'(1x,i5,1x,f10.3,1x,f11.5)') &
                ib, ww(ib,1), ee(ib,1)*tohart
        end do
     else
        do ib=npbb,npbe
           write(6,'(1x,i5,1x,f10.3,1x,f11.5,1x,f8.5,1x,f11.5)') &
                ib, ww(ib,1), ee(ib,1)*tohart, ww(ib,2), ee(ib,2)*tohart
        end do
     end if

     write(6,*)
  end do

  if (nspin.eq.1) then
     write(6,'(1x,a,i5)') '# index, min(eig energy),' &
     // ' max(eig energy), total_weight'
     do ib=npbb,npbe
        write(6,'(1x,i5,3(1x,f13.5))') &
             ib, eemin(ib,1)*tohart, eemax(ib,1)*tohart, wsum(ib,1)
     end do
  else
     write(6,'(1x,a,i5)') '# index, each_spin[ min(eig energy),' &
     // ' max(eig energy) ]'
     do ib=npbb,npbe
        write(6,'(1x,i5,6(1x,f13.5))') &
             ib, eemin(ib,1)*tohart, eemax(ib,1)*tohart, wsum(ib,1), &
             eemin(ib,2)*tohart, eemax(ib,2)*tohart, wsum(ib,2)
     end do
  end if
  write(6,*)

  if (nspin.eq.1) then
     write(6,'(1x,a,i5)') '# index, gap(eig energy)'
     do ib=max(2,npbb),npbe
        write(6,'(2(1x,i5),1x,f13.5)') &
             ib-1, ib, ( eemin(ib,1)-eemax(ib-1,1) )*tohart
     end do
  else
     write(6,'(1x,a,i5)') '# index, each_spin[ gap(eig energy) ]'
     do ib=max(2,npbb),npbe
        write(6,'(2(1x,i5),2(1x,f13.5))') &
             ib-1, ib, ( eemin(ib,1)-eemax(ib-1,1) )*tohart, &
             ( eemin(ib,2)-eemax(ib-1,2) )*tohart
     end do
  end if

end program wfneig
